O_DEFAULT = '@'
O_NATIVE = '='
O_LITTLE = '<'
O_BIG = '>'
O_NETWORK = '!'

Pad = 'x'
Bool = '?'
Char = 'c'
Int1 = 'b'
uInt1 = 'B'
Int2 = 'h'
uInt2 = 'H'
Int4 = 'l'
uInt4 = 'L'
Int8 = 'q'
uInt8 = 'Q'
Float2 = 'e'
Float4 = 'f'
Float8 = 'd'
String = 's'
PascalString = 'p'
